from PyQt5 import QtCore, QtGui, QtWidgets
from PyQt5.QtWidgets import QFileDialog, QMessageBox, qApp, QStyle
import config
import numpy as np
import os
import pbar
import rfl
import rflGui # Created by designer.
import sys

'''
This is a messy file that simply glues rflGui.ui widgets to corresponding
rfl subroutines.  It is not complicated but was tedious to create.  I
discovered that I do not enjoy making GUIs.  :-)

Mike Markowski, mike.ab3ap@gmail.com
Apr 2023
'''

class GuiConnect:

	def __init__(self):
		global d, u
		d = None # Dictionary from config file.
		u = None # Short name.

	def dictToGui(self):
		global d, u

		# Top of panel.
		if 'mode' in d['rfl'].keys():
			if d['rfl']['mode'] == 'profile':
				u.rbProfile.setChecked(True)
				u.cbSiting.setEnabled(False)
				self.handleRbProfile()
			else: # Area.
				u.rbArea.setChecked(True)
				u.cbSiting.setEnabled(True)
				self.handleRbArea()
		if 'output' in d['rfl'].keys():
			if d['rfl']['output'] == 'coverage':
				u.rbCoverage.setChecked(True)
				self.handleRbCoverage()
			else: # Link.
				u.rbLink.setChecked(True)
				self.handleRbCoverage()

		self.dictRflToGui()
		self.dictCoverageToGui()
		self.dictEnvToGui()
		self.dictNoiseToGui()
		self.dictStationToGui()

	def dictCoverageToGui(self):
		global d, u
		if 'grid' in d['coverage'].keys():
			u.leGrid.setText(str(d['coverage']['grid']))
		if 'range' in d['coverage'].keys():
			u.leRange.setText(str(d['coverage']['range']))
		if 'lat' in d['coverage'].keys() and d['coverage']['lat'] != '':
			u.leLatCtr.setText('%.6f' % np.degrees(d['coverage']['lat']))
		if 'lon' in d['coverage'].keys() and d['coverage']['lon'] != '':
			u.leLonCtr.setText('%.6f' % np.degrees(d['coverage']['lon']))
		if 'contour' in d['coverage'].keys():
			if d['coverage']['contour'] == 'fill':
				u.rbFilled.setChecked(True)
			else:
				u.rbLines.setChecked(True)
		if 'pattern' in d['coverage'].keys():
			if d['coverage']['pattern'] == 'yes':
				u.rbAntYes.setChecked(True)
			else:
				u.rbAntNo.setChecked(True)
		if 'title' in d['coverage'].keys():
			u.leMap.setText(str(d['coverage']['title']))
		if 'levels' in d['coverage'].keys():
			s = str(d['coverage']['levels'])
			s = s[1:-1] # Remove square brackets.
			s = s.split() # Remove white space.
			s = ''.join(s)
			u.leLevels.setText(s)

	def dictEnvToGui(self):
		global d, u
		# env tab, multiple EM environments can exist.
		for e in d['env'].keys(): # Loop through each env.
			u.cbAllEnv.addItem(e)
			i = u.cbAllEnv.findText(e)
			u.cbAllEnv.setCurrentIndex(i)
			u.cbStationEnv.addItem(e)
			u.leEnvName.setText(e)
			u.leEps.setText(str(d['env'][e]['eps']))
			u.leSigma.setText(str(d['env'][e]['sigma']))
			u.leN0.setText(str(d['env'][e]['N0']))

	def dictNoiseToGui(self):
		global d, u
		# noise tab, multiple noise environments can exist.
		for n in d['noise'].keys(): # Loop through each noise env.
			u.cbAllNoise.addItem(n)
			u.cbStationNoise.addItem(n)
			u.leNoiseName.setText(n)
			u.leBw.setText(str(d['noise'][n]['B']))
			u.leFreq.setText(str(d['noise'][n]['f']))
			u.leNf.setText(str(d['noise'][n]['NF']))
			u.leTemp.setText(str(d['noise'][n]['T']))
			s = d['noise'][n]['env']
			i = u.cbEnv.findText(s)
			u.cbEnv.setCurrentIndex(i)

	def dictRflToGui(self):
		global d, u
		if 'ant' in d['rfl'].keys():
			u.leAntenna.setText(d['rfl']['ant'])
		if 'terrain' in d['rfl'].keys():
			s = d['rfl']['terrain']
			s = 'DTED' if s=='dted' else 'Open Elev'
			i = u.cbTerrain.findText(s)
			u.cbTerrain.setCurrentIndex(i)
		if 'dted' in d['rfl'].keys():
			u.leDted.setText(d['rfl']['dted'])
		if 'outfile' in d['rfl'].keys():
			u.leOutfile.setText(d['rfl']['outfile'])
		if 'units' in d['rfl'].keys():
			s = d['rfl']['units']
			s = 'S/N dB' if s=='SNR' else s
			i = u.cbUnits.findText(s)
			u.cbUnits.setCurrentIndex(i)
			u.lUnits.setText(s)
		if 'conf' in d['rfl'].keys():
			u.leConf.setText(str(d['rfl']['conf']))
		if 'time' in d['rfl'].keys():
			u.leTime.setText(str(d['rfl']['time']))
		if 'loc' in d['rfl'].keys():
			u.leLoc.setText(str(d['rfl']['loc']))
		if 'dh' in d['rfl'].keys():
			u.leDh.setText(str(d['rfl']['dh']))
		if ('interference' in d['rfl'].keys()
			and d['rfl']['interference'] == 'yes'):
			u.rbIntYes.setChecked(True)
		else:
			d['rfl']['interference'] == 'no'
			u.rbIntNo.setChecked(True)

	def dictStationToGui(self):
		global d, u
		# station tab, multiple stations can exist.
		for s in d['station'].keys(): # Loop through each station.
			u.cbAllStations.addItem(s)
			i = u.cbAllStations.findText(s)
			u.cbAllStations.setCurrentIndex(i)
			u.leStationName.setText(s)
			self.guiLoadStation(s)

	def dialogBox(self, msgType, title, info, parent=None):
		match msgType:
			case '?':
				it = QMessageBox.Question
			case 'i':
				it = QMessageBox.Information
			case 'w':
				it = QMessageBox.Warning
			case 'c':
				it = QMessageBox.Critical
			case _:
				it = QMessageBox.NoIcon
		b = QMessageBox(parent)
		b.setIcon(it)
		b.setText(title)
		b.setInformativeText(info)
		b.setStandardButtons(QMessageBox.Ok | QMessageBox.Cancel)
		b.setDefaultButton(QMessageBox.Cancel)
		return b.exec()

	def errorBox(self, msgType, title, info, parent=None):
		match msgType:
			case '?':
				it = QMessageBox.Question
			case 'i':
				it = QMessageBox.Information
			case 'w':
				it = QMessageBox.Warning
			case 'c':
				it = QMessageBox.Critical
			case _:
				it = QMessageBox.NoIcon
		b = QMessageBox(parent)
		b.setIcon(it)
		b.setText(title)
		b.setInformativeText(info)
		b.setStandardButtons(QMessageBox.Ok)
		b.exec()
		err = False

	def gui(self):
		global d, u

		app = QtWidgets.QApplication(sys.argv)
		rfl = QtWidgets.QMainWindow()
		u = rflGui.Ui_rfl()
		u.setupUi(rfl)

		# Main panel widgets.
		u.rbProfile.toggled.connect(self.handleRbProfile)
		self.handleRbProfile()
		u.rbArea.toggled.connect(self.handleRbArea)

		u.rbCoverage.toggled.connect(self.handleRbCoverage)
		u.rbLink.toggled.connect(self.handleRbLink)
		u.pbRead.clicked.connect(self.handlePbRead)
		u.pbWrite.clicked.connect(self.handlePbWrite)
		u.pbRun.clicked.connect(self.handlePbRun)

		# rfl tab
		u.pbAntenna.clicked.connect(self.handlePbAnt)
		u.pbDted.clicked.connect(self.handlePbDted)
		u.cbTerrain.currentTextChanged.connect(self.handleCbTerrain)
		u.leConf.editingFinished.connect(self.handleLeConf)
		u.leTime.editingFinished.connect(self.handleLeTime)
		u.leLoc.editingFinished.connect(self.handleLeLoc)
		u.leDh.editingFinished.connect(self.handleLeDh)
		u.pbRCancel.clicked.connect(self.dictRflToGui)
		u.pbRUpdate.clicked.connect(self.handlePbRUpdate)

		# coverage tab
		u.leGrid.editingFinished.connect(self.handleLeGrid)
		u.leRange.editingFinished.connect(self.handleLeRange)
		u.pbCCancel.clicked.connect(self.dictCoverageToGui)
		u.pbCUpdate.clicked.connect(self.handlePbCUpdate)

		# env tab.
		u.cbAllEnv.currentTextChanged.connect(self.handleCbAllEnv)
		u.leEps.editingFinished.connect(self.handleLeEps)
		u.leSigma.editingFinished.connect(self.handleLeSigma)
		u.leN0.editingFinished.connect(self.handleLeN0)
		trashIcon = qApp.style().standardIcon(QStyle.SP_TrashIcon)
		u.eDelete.setIcon(trashIcon)
		u.eDelete.clicked.connect(self.handlePbEDelete)
		u.pbECancel.clicked.connect(self.dictEnvToGui)
		u.pbEUpdate.clicked.connect(self.handlePbEUpdate)

		# noise tab.
		u.cbAllNoise.currentTextChanged.connect(self.handleCbAllNoise)
		u.leBw.editingFinished.connect(self.handleLeBw)
		u.leFreq.editingFinished.connect(self.handleLeFreq)
		u.leNf.editingFinished.connect(self.handleLeNf)
		u.leTemp.editingFinished.connect(self.handleLeTemp)
		u.nDelete.setIcon(trashIcon)
		u.nDelete.clicked.connect(self.handlePbNDelete)
		u.pbNCancel.clicked.connect(self.dictNoiseToGui)
		u.pbNUpdate.clicked.connect(self.handlePbNUpdate)

		# station tab.
		u.cbAllStations.currentTextChanged.connect(self.handleCbAllStations)
		u.rbTx.toggled.connect(self.handleRbTx)
		u.rbRx.toggled.connect(self.handleRbRx)
		u.cbNsma.currentTextChanged.connect(self.handleCbNsma)
		u.pbNsma.clicked.connect(self.handlePbNsma)
		u.leAz.editingFinished.connect(self.handleLeAz)
		u.leEl.editingFinished.connect(self.handleLeEl)
		u.leHgt.editingFinished.connect(self.handleLeHgt)
		u.leLat.editingFinished.connect(self.handleLeLat)
		u.leLon.editingFinished.connect(self.handleLeLon)
		u.leSys.editingFinished.connect(self.handleLeSys)
		u.leTx.editingFinished.connect(self.handleLeTx)
		u.cbSiting.currentTextChanged.connect(self.handleCbSiting)
		u.sDelete.setIcon(trashIcon)
		u.sDelete.clicked.connect(self.handlePbSDelete)
		u.pbSCancel.clicked.connect(self.dictStationToGui)
		u.pbSUpdate.clicked.connect(self.handlePbSUpdate)

		d = config.cfgEmpty()
		d['rfl']['output'] = 'coverage'
		d['rfl']['mode'] = 'profile'
		self.dictToGui()

		rfl.show()
		sys.exit(app.exec_())

	def guiLoadEnv(self, e):
		global d, u

		i = u.cbAllEnv.findText(e)
		if i == -1:
			return
		u.leEnvName.setText(e)
		if 'eps' in d['env'].keys():
			u.leEps.setText('%.3f' % d['env'][e]['eps'])
		if 'env' in d['env'].keys():
			u.leSigma.setText('%.3f' % d['env'][e]['sigma'])
		if 'N0' in d['env'].keys():
			u.leN0.setText('%.3f' % d['env'][e]['N0'])

	def guiLoadNoise(self, n):
		global d, u

		i = u.cbAllNoise.findText(n)
		if i == -1:
			return
		u.leNoiseName.setText(n)
		if 'B' in d['noise'].keys():
			u.leBw.setText('%.3f' % d['noise'][n]['B'])
		if 'f' in d['noise'].keys():
			u.leFreq.setText('%s' % d['noise'][n]['f'])
		if 'NF' in d['noise'].keys():
			u.leNf.setText('%.1f' % d['noise'][n]['NF'])
		if 'T' in d['noise'].keys():
			u.leTemp.setText('%.1f' % d['noise'][n]['T'])
			i = u.cbEnv.findText(d['noise'][n]['env'])
			u.cbEnv.setCurrentIndex(i)

	def guiLoadStation(self, s):
		global d, u

		i = u.cbAllStations.findText(s)
		if i == -1:
			return
		u.leStationName.setText(s)
		if d['station'][s]['ant'].strip() != '':
			a = d['station'][s]['ant']
			u.leNsma.setText(a)
			try:
				float(a)
				i = u.cbNsma.findText('dBi')
			except ValueError:
				i = u.cbNsma.findText('NSMA')
			u.cbNsma.setCurrentIndex(i)
			self.handleCbNsma()
		if d['station'][s]['pol'].strip() != '':
			p = d['station'][s]['pol'].lower()
			if p == 'h':
				i = u.cbPol.findText('Horizontal')
			else:
				i = u.cbPol.findText('Vertical')
			u.cbPol.setCurrentIndex(i)
		n = self.findFloat(d['station'][s]['az'])
		if n != '':
			u.leAz.setText('%.1f' % np.degrees(n))
		n = self.findFloat(d['station'][s]['el'])
		if n != '':
			u.leEl.setText('%.1f' % np.degrees(n))
		n = self.findFloat(d['station'][s]['hgt'])
		if n != '':
			u.leHgt.setText('%.1f' % n)
		if 'lat' in d['station'][s].keys() and d['station'][s]['lat'] != '':
			u.leLat.setText('%.6f' % np.degrees(d['station'][s]['lat']))
		else:
			u.leLat.setText('')
		if 'lon' in d['station'][s].keys() and d['station'][s]['lon'] != '':
			u.leLon.setText('%.6f' % np.degrees(d['station'][s]['lon']))
		else:
			u.leLon.setText('')
		i = u.cbStationEnv.findText(d['station'][s]['env'])
		u.cbStationEnv.setCurrentIndex(i)
		i = u.cbStationNoise.findText(d['station'][s]['noise'])
		u.cbStationNoise.setCurrentIndex(i)
		u.leSys.setText('%.1f' % d['station'][s]['sys'])
		if 'tx' in d['station'][s].keys():
			u.leTx.setText('%.1f' % d['station'][s]['tx'])
			u.rbTx.setChecked(True)
			if not 'service' in d['station'][s].keys():
				d['station'][s]['service'] = 3
			u.cbService.setCurrentIndex(d['station'][s]['service'])
		else:
			u.leTx.setText('')
			u.rbRx.setChecked(True)
		if d['station'][s]['site'].strip() != '':
			match d['station'][s]['site'].lower():
				case 'r':
					i = u.cbSiting.findText('Random')
				case 'c':
					i = u.cbSiting.findText('Careful')
				case 'v':
					i = u.cbSiting.findText('Very Careful')
			u.cbSiting.setCurrentIndex(i)

	def handleCbAllEnv(self):
		global d, u
		s = u.cbAllEnv.currentText() # Selected EM environment.
		self.guiLoadEnv(s)

	def handleCbAllNoise(self):
		global d, u
		n = u.cbAllNoise.currentText() # Selected noise environment.
		self.guiLoadNoise(n)

	def handleCbAllStations(self):
		global d, u
		s = u.cbAllStations.currentText() # Newly selected station.
		self.guiLoadStation(s)

	def handleCbNsma(self):
		global u
		u.pbNsma.setEnabled(u.cbNsma.currentText() == 'NSMA')

	def handleCbSiting(self):
		global d, u
		stn = u.leStationName.text().strip()
		if stn == '':
			return
		site = u.cbSiting.currentText()
		d['station'][stn]['site'] = site

	def handleCbTerrain(self):
		global d, u
		terrain = u.cbTerrain.currentText()
		if terrain == 'DTED':
			u.leDted.setText('')
		else:
			u.leDted.setText('https://api.open-elevation.com/api/v1/lookup')

	def handleLeAz(self):
		global d, u
		sName = u.leStationName.text().strip()
		if sName == '':
			self.errorBox('i', '<b>Bad Station Name</b>',
				'Choose or add name.', parent=u.tabWidget)
			u.leAz.setText('')
			return
		az = u.leAz.text().strip()
		if az == '':
			return
		try:
			float(u.leAz.text())
		except ValueError:
			self.errorBox('c', '<b>Bad Number</b>',
				'Azimuth must be a number.', parent=u.tabWidget)

	def handleLeBw(self):
		global d, u
		nName = u.leNoiseName.text().strip()
		if nName == '':
			self.errorBox('i', '<b>Bad Noise Name</b>',
				'Choose or add name.', parent=u.tabWidget)
			u.leBw.setText('')
			return
		bw = u.leBw.text().strip()
		if bw == '':
			return
		try:
			float(u.leBw.text().strip())
		except ValueError:
			self.errorBox('c', '<b>Bad Number</b>',
				'Bandwidth must be a number.', parent=u.tabWidget)

	def handleLeConf(self):
		global d, u
		try:
			float(u.leConf.text().strip())
		except ValueError:
			self.errorBox('c', '<b>Bad Confidence</b>',
				'Confidence must be 0 to 100.', parent=u.tabWidget)

	def handleLeDh(self):
		global d, u
		try:
			float(u.leDh.text().strip())
		except ValueError:
			self.errorBox('c', '<b>Bad Delta h</b>',
				'Delta-h must be a number.', parent=u.tabWidget)

	def handleLeEl(self):
		global d, u
		sName = u.leStationName.text().strip()
		if sName == '':
			self.errorBox('i', '<b>Bad Station Name</b>',
				'Choose or add name.', parent=u.tabWidget)
			u.leEl.setText('')
			return
		el = u.leEl.text().strip()
		if el == '':
			return
		try:
			float(u.leEl.text())
		except ValueError:
			self.errorBox('c', '<b>Bad Number</b>',
				'Azimuth must be a number.', parent=u.tabWidget)

	def handleLeEps(self):
		global d, u
		eName = u.leEnvName.text().strip()
		if eName == '':
			self.errorBox('i', '<b>Bad Environment Name</b>',
				'Choose or add name.', parent=u.tabWidget)
			u.leEps.setText('')
			return
		eps = u.leEps.text().strip()
		if eps == '':
			return
		try:
			float(u.leEps.text().strip())
		except ValueError:
			self.errorBox('c', '<b>Bad Number</b>',
				'Epsilon must be a number.', parent=u.tabWidget)

	def handleLeFreq(self):
		global d, u
		nName = u.leNoiseName.text().strip()
		if nName == '':
			self.errorBox('i', '<b>Bad Noise Name</b>',
				'Choose or add name.', parent=u.tabWidget)
			u.leFreq.setText('')
			return
		freq = u.leFreq.text().strip()
		if freq == '':
			return
		try:
			float(u.leFreq.text().strip())
		except ValueError:
			self.errorBox('c', '<b>Bad Number</b>',
				'Bandwidth must be a number.', parent=u.tabWidget)

	def handleLeHgt(self):
		global d, u
		sName = u.leStationName.text().strip()
		if sName == '':
			self.errorBox('i', '<b>Bad Station Name</b>',
				'Choose or add name.', parent=u.tabWidget)
			u.leHgt.setText('')
			return
		hgt = u.leHgt.text().strip()
		if hgt == '':
			return
		try:
			float(u.leHgt.text().strip())
		except ValueError:
			self.errorBox('c', '<b>Bad Number</b>',
				'Height must be a number.', parent=u.tabWidget)

	def handleLeLat(self):
		global d, u
		sName = u.leStationName.text().strip()
		if sName == '':
			self.errorBox('i', '<b>Bad Station Name</b>',
				'Choose or add name.', parent=u.tabWidget)
			u.leLat.setText('')
			return
		lat = u.leLat.text().strip()
		if lat == '':
			return
		try:
			np.radians(float(u.leLat.text()))
		except ValueError:
			self.errorBox('c', '<b>Bad Number</b>',
				'Latitude must be decimal degrees.', parent=u.tabWidget)

	def handleLeLon(self):
		global d, u
		sName = u.leStationName.text().strip()
		if sName == '':
			self.errorBox('i', '<b>Bad Station Name</b>',
				'Choose or add name.', parent=u.tabWidget)
			u.leLon.setText('')
			return
		lon = u.leLon.text().strip()
		if lon == '':
			return
		try:
			float(u.leLon.text())
		except ValueError:
			self.errorBox('c', '<b>Bad Number</b>',
				'Longitude must be decimal degrees.', parent=u.tabWidget)

	def handleLeN0(self):
		global d, u
		eName = u.leEnvName.text().strip()
		if eName == '':
			self.errorBox('i', '<b>Bad Environment Name</b>',
				'Choose or add name.', parent=u.tabWidget)
			u.leN0.setText('')
			return
		N0 = u.leN0.text().strip()
		if N0 == '':
			return
		try:
			float(u.leN0.text().strip())
		except ValueError:
			self.errorBox('c', '<b>Bad Number</b>',
				'N0 must be a number.', parent=u.tabWidget)

	def handleLeNf(self):
		global d, u
		nName = u.leNoiseName.text().strip()
		if nName == '':
			self.errorBox('i', '<b>Bad Noise Name</b>',
				'Choose or add name.', parent=u.tabWidget)
			u.leNf.setText('')
			return
		nf = u.leNf.text().strip()
		if nf == '':
			return
		try:
			float(u.leNf.text())
		except ValueError:
			self.errorBox('c', '<b>Bad Number</b>',
				'Noise figure must be a number.', parent=u.tabWidget)

	def handleLeGrid(self):
		global d, u
		try:
			int(u.leGrid.text())
		except ValueError:
			self.errorBox('c', '<b>Bad Integer</b>',
				'Grid size must be an integer.', parent=u.tabWidget)

	def handleLeLoc(self):
		global d, u
		try:
			float(u.leLoc.text())
		except ValueError:
			self.errorBox('c', '<b>Bad Loc</b>',
				'Location Fraction must be 0 to 100.', parent=u.tabWidget)

	def handleLeRange(self):
		global d, u
		try:
			float(u.leRange.text())
		except ValueError:
			self.errorBox('c', '<b>Bad Range</b>',
				'Range must be a number.', parent=u.tabWidget)

	def handleLeSigma(self):
		global d, u
		eName = u.leEnvName.text().strip()
		if eName == '':
			self.errorBox('i', '<b>Bad Environment Name</b>',
				'Choose or add name.', parent=u.tabWidget)
			u.leSigma.setText('')
			return
		sigma = u.leSigma.text().strip()
		if sigma == '':
			return
		try:
			float(sigma)
		except ValueError:
			self.errorBox('c', '<b>Bad Number</b>',
				'Sigma must be a number.', parent=u.tabWidget)

	def handleLeSys(self):
		global d, u
		sName = u.leStationName.text().strip()
		if sName == '':
			self.errorBox('i', '<b>Bad Station Name</b>',
				'Choose or add name.', parent=u.tabWidget)
			u.leAz.setText('')
			return
		sys = u.leSys.text().strip()
		if sys == '':
			return
		try:
			float(u.leSys.text().strip())
		except ValueError:
			self.errorBox('c', '<b>Bad Number</b>',
				'Loss must be a number.', parent=u.tabWidget)

	def handleLeTx(self):
		global d, u
		sName = u.leStationName.text().strip()
		if sName == '':
			self.errorBox('i', '<b>Bad Station Name</b>',
				'Choose or add name.', parent=u.tabWidget)
			u.leAz.setText('')
			return
		tx = u.leTx.text().strip()
		if tx == '':
			return
		try:
			float(u.leTx.text().strip())
		except ValueError:
			self.errorBox('c', '<b>Bad Number</b>',
				'Power must be a number.', parent=u.tabWidget)

	def handleLeTemp(self):
		global d, u
		nName = u.leNoiseName.text().strip()
		if nName == '':
			self.errorBox('i', '<b>Bad Noise Name</b>',
				'Choose or add name.', parent=u.tabWidget)
			u.leTemp.setText('')
			return
		temp = u.leTemp.text().strip()
		if temp == '':
			return
		try:
			float(u.leTemp.text().strip())
		except ValueError:
			self.errorBox('c', '<b>Bad Number</b>',
				'Noise figure must be a number.', parent=u.tabWidget)

	def handleLeTime(self):
		global d, u
		try:
			float(u.leTime.text())
		except ValueError:
			self.errorBox('c', '<b>Bad Time</b>',
				'Time fraction must be 0 to 100.', parent=u.tabWidget)

	def handlePbAnt(self):
		global u
		dirPath = QtWidgets.QFileDialog.getExistingDirectory(
			u.tabWidget, 'Select Antenna Directory')
		u.leAntenna.setText(dirPath)

	def handlePbCUpdate(self):
		global d, u
		d['coverage']['grid'] = self.findInt(u.leGrid.text())
		d['coverage']['range'] = self.findFloat(u.leRange.text())
		d['coverage']['title'] = u.leMap.text().strip()
		d['coverage']['pattern'] = 'yes' if u.rbAntYes.isChecked() else 'no'
		d['coverage']['contour'] = 'fill' if u.rbFilled.isChecked() else 'line'
		if u.leLevels.text().strip() != '':
			lvl = u.leLevels.text().split(',') # Remove white space.
			val = list(map(float,lvl)) # List of floats.
			d['coverage']['levels'] = val
		elif 'levels' in d['coverage'].keys():
			del d['coverage']['levels']
		if u.leLatCtr.text().strip() != '':
			d['coverage']['lat'] = np.radians(float(u.leLatCtr.text()))
		if u.leLonCtr.text().strip() != '':
			d['coverage']['lon'] = np.radians(float(u.leLonCtr.text()))

	def handlePbDted(self):
		global u
		dirPath = QtWidgets.QFileDialog.getExistingDirectory(
			u.tabWidget, 'Select DTED Directory')
		u.leDted.setText(dirPath)

	def handlePbEDelete(self):
		global u
		i = u.cbAllEnv.currentIndex()
		if i == -1:
			return
		nEnv = u.cbAllEnv.currentText()
		q = 'Really Delete %s?' % nEnv
		ret = self.dialogBox('w', '<b>Noise Delete Warning</b>', q,
			parent=u.tabWidget)
		if ret == QMessageBox.Cancel:
			return
		del d['env'][nEnv]
		u.cbAllEnv.removeItem(i)
		i = u.cbStationEnv.currentIndex()
		u.cbStationEnv.removeItem(i)

	def handlePbNDelete(self):
		global d, u
		i = u.cbAllNoise.currentIndex()
		if i == -1:
			return
		nNoise = u.cbAllNoise.currentText()
		q = 'Really Delete %s?' % nNoise
		ret = self.dialogBox('w', '<b>Noise Delete Warning</b>', q,
			parent=u.tabWidget)
		if ret == QMessageBox.Cancel:
			return
		del d['noise'][nNoise]
		u.cbAllNoise.removeItem(i)
		i = u.cbStationNoise.currentIndex()
		u.cbStationNoise.removeItem(i)

	def handlePbSDelete(self):
		global d, u
		i = u.cbAllStations.currentIndex()
		if i == -1:
			return
		stn = u.cbAllStations.currentText()
		q = 'Really Delete %s?' % stn
		ret = self.dialogBox('w', '<b>Station Delete Warning</b>', q,
			parent=u.tabWidget)
		if ret == QMessageBox.Cancel:
			return
		del d['station'][stn]
		u.cbAllStations.removeItem(i)

	def findFloat(self, s, toRadians=False, ifBlank=0):
		if not type(s) is str:
			return s
		if s.strip() == '':
			return ifBlank
		n = float(s)
		return np.radians(n) if toRadians else float(s)

	def findInt(self, s):
		if not type(s) is str:
			return s
		return 0 if s.strip()=='' else int(s)

	def handlePbEUpdate(self):
		global d, u
		eName = u.leEnvName.text().strip()
		envNew = -1 == u.cbAllEnv.findText(eName)
		if envNew:
			d['env'][eName] = config.cfgEmptyEnv()
		d['env'][eName]['eps'] = self.findFloat(u.leEps.text())
		d['env'][eName]['sigma'] = self.findFloat(u.leSigma.text())
		d['env'][eName]['N0'] = self.findFloat(u.leN0.text())
		d['env'][eName]['climate'] = u.cbClimate.currentIndex()
		if envNew:
			u.cbAllEnv.addItem(eName)
			u.cbStationEnv.addItem(eName)
			i = u.cbAllEnv.findText(eName)
			u.cbAllEnv.setCurrentIndex(i)

	def handlePbNsma(self):
		global u
		filetype='(*.nsma);;(*.adf);;All files (*)'
		u.pbNsma.setEnabled(u.cbNsma.currentText() == 'NSMA')
		filename, filetype = QFileDialog.getOpenFileName(
			u.tabWidget, 'NSMA File Chooser', '.', filetype)
		u.leNsma.setText(filename)

	def handlePbNUpdate(self):
		global d, u
		nName = u.leNoiseName.text().strip()
		noiseNew = -1 == u.cbAllNoise.findText(nName)
		if noiseNew:
			d['noise'][nName] = config.cfgEmptyNoise()
		d['noise'][nName]['B'] = self.findFloat(u.leBw.text())
		d['noise'][nName]['env'] = u.cbEnv.currentText()
		d['noise'][nName]['f'] = u.leFreq.text() # Must remain string.
		d['noise'][nName]['NF'] = self.findFloat(u.leNf.text())
		d['noise'][nName]['T'] = self.findFloat(u.leTemp.text())
		if noiseNew:
			u.cbAllNoise.addItem(nName)
			u.cbStationNoise.addItem(nName)
			i = u.cbAllNoise.findText(nName)
			u.cbAllNoise.setCurrentIndex(i)

	def handlePbRead(self):
		global d, u

		# Open file browser.
		filetype='rfl config (*.rfl);;All files (*)'
		filename, filetype = QFileDialog.getOpenFileName(
			u.tabWidget, 'RFL Config Chooser', '.', filetype)

		# Retrieve filename user chose.
		u.leConfig.setText(filename)
		filename = u.leConfig.text().strip()
		if filename == '':
			self.errorBox('c', '<b>No Configuration</b>',
				'Choose configuration file.', parent=u.tabWidget)
			return

		# Open file, clear old GUI values, and populate GUI.
		d = None
		d = config.cfgParse(config.cfgRead(filename))
		u.cbAllNoise.clear()
		u.cbStationNoise.clear()
		u.cbAllEnv.clear()
		u.cbStationEnv.clear()
		u.cbAllStations.clear()
		u.cbAllNoise.clear()
		u.cbAllEnv.clear()
		self.dictToGui()

	def handlePbRun(self):
		global d, u
		if d is None:
			return
		p = pbar.PBar(u.tabWidget)
		rfl.progress = p.progress
		rfl.rfPredict(d, gui=True)
		p.close()

	def handlePbRUpdate(self):
		global d, u
		d['rfl']['ant'] = u.leAntenna.text().strip()
		s = u.cbTerrain.currentText().strip()
		s = 'dted' if s=='DTED' else 'open'
		d['rfl']['terrain'] = s
		d['rfl']['dted'] = u.leDted.text().strip()

		fname = u.leOutfile.text().strip()
		if fname == '':
			fname = 'rfl'
		if u.rbCoverage.isChecked():
			if fname.find('.png') == -1:
				fname += '.png'
		else: # Link mode, csv file wanted.
			if fname.find('.csv') == -1:
				fname += '.csv'

		d['rfl']['outfile'] = fname
		d['rfl']['conf'] = self.findFloat(u.leConf.text(), ifBlank='')
		d['rfl']['time'] = self.findFloat(u.leTime.text(), ifBlank='')
		d['rfl']['loc'] = self.findFloat(u.leLoc.text(), ifBlank='')
		d['rfl']['dh'] = self.findFloat(u.leDh.text(), ifBlank='')
		d['rfl']['interference'] = 'yes' if u.rbIntYes.isChecked() else 'no'

		units = u.cbUnits.currentText()
		units = 'SNR' if units=='S/N dB' else units
		u.lUnits.setText(units)
		d['rfl']['units'] = units

	def handlePbSUpdate(self):
		global d, u
		sName = u.leStationName.text().strip()
		stnNew = -1 == u.cbAllStations.findText(sName)
		if stnNew:
			d['station'][sName] = config.cfgEmptyStation()
			i = u.cbAllStations.findText(sName)
			u.cbAllStations.setCurrentIndex(i)
		d['station'][sName]['service'] = u.cbService.currentIndex()

		d['station'][sName]['ant'] = u.leNsma.text().strip()
		if u.cbPol.currentIndex() != -1:
			pol = u.cbPol.currentText()
			pol = 'v' if pol=='Vertical' else 'h'
			d['station'][sName]['pol'] = pol
		d['station'][sName]['az'] = self.findFloat(u.leAz.text(), True)
		d['station'][sName]['el'] = self.findFloat(u.leEl.text(), True)
		d['station'][sName]['hgt'] = self.findFloat(u.leHgt.text())
		d['station'][sName]['lat'] = self.findFloat(u.leLat.text(), True, '')
		d['station'][sName]['lon'] = self.findFloat(u.leLon.text(), True, '')
		d['station'][sName]['env'] = u.cbStationEnv.currentText()
		d['station'][sName]['noise'] = u.cbStationNoise.currentText()
		v = self.findFloat(u.leSys.text())
		v = 0 if v=='' else v
		d['station'][sName]['sys'] = v
		if u.leTx.text().strip() != '':
			d['station'][sName]['tx'] = self.findFloat(u.leTx.text())
		d['station'][sName]['env'] = u.cbStationEnv.currentText()
		d['station'][sName]['noise'] = u.cbStationNoise.currentText()
		d['station'][sName]['site'] = u.cbSiting.currentText()

		site = u.cbSiting.currentText()
		match site:
			case 'Random':
				site = 'r'
			case 'Careful':
				site = 'c'
			case 'Very Careful':
				site = 'v'
		d['station'][sName]['site'] = site
		if stnNew:
			u.cbAllStations.addItem(sName)
			i = u.cbAllStations.findText(sName)

	def handlePbWrite(self):
		global d, u
		fname = u.leConfig.text()
		if fname == '':
			self.errorBox('c', '<b>No Output File</b>',
				'Choose output file.', parent=u.tabWidget)
			return
		if os.path.exists(fname):
			ret = self.dialogBox('w', '<b>File Exists</b>', 'Overwrite?',
				parent=u.tabWidget)
			if ret == QMessageBox.Cancel:
				return

		f = open(fname, 'w')

		# Write out rfl block.
		r = d['rfl']
		f.write('model rfl\n')
		if self.isWord(r, 'ant'):
			f.write('  ant %s\n' % r['ant'])
		if self.isWord(r, 'terrain'):
			f.write('  terrain %s\n' % r['terrain'])
		if self.isWord(r, 'dted'):
			f.write('  dted %s\n' % r['dted'])
		if self.isWord(r, 'units'):
			f.write('  units %s\n' % r['units'])
		if self.isWord(r, 'mode'):
			f.write('  mode %s\n' % r['mode'])
		if self.isNumber(r, 'dh'):
			f.write('  dh %f\n' % r['dh'])
		if self.isWord(r, 'output'):
			f.write('  output %s\n' % r['output'])
		if self.isWord(r, 'outfile'):
			f.write('  outfile %s\n' % r['outfile'])
		if self.isNumber(r,'conf'):
			f.write('  conf %f\n' % r['conf'])
		if self.isNumber(r, 'time'):
			f.write('  time %f\n' % r['time'])
		if self.isNumber(r, 'loc'):
			f.write('  loc %f\n' % r['loc'])
		f.write('  interference %s\n' % d['rfl']['interference'])

		# Write out coverage block.
		if 'coverage' in d.keys():
			f.write('\nmodel coverage\n')
			c = d['coverage']
			if self.isNumber(c, 'grid'):
				f.write('  grid %d\n' % c['grid'])
			if self.isNumber(c, 'range'):
				f.write('  range %f\n' % c['range'])
			if self.isNumber(c, 'lat'):
				f.write('  lat %f\n' % np.degrees(c['lat']))
			if self.isNumber(c, 'lon'):
				f.write('  lon %f\n' % np.degrees(c['lon']))
			if self.isWord(c, 'title'):
				f.write('  title %s\n' % c['title'])
			if self.isWord(c, 'levels'):
				s = str(c['levels'])
				s = s.replace('[', '')
				s = s.replace(']', '')
				s = s.split() # Remove white space.
				s = ''.join(s)
				f.write('  levels %s\n' % s)
			if self.isWord(c, 'pattern'):
				f.write('  pattern %s\n' % c['pattern'])
			if self.isWord(c, 'contour'):
				f.write('  contour %s\n' % c['contour'])

		# Write out env blocks.
		for eName in d['env'].keys():
			e = d['env'][eName]
			f.write('\nmodel env %s\n' % eName)
			f.write('  climate %d\n' % e['climate'])
			if self.isNumber(e, 'eps'):
				f.write('  eps %f\n' % e['eps'])
			if self.isNumber(e, 'N0'):
				f.write('  N0 %f\n' % e['N0'])
			if self.isNumber(e, 'sigma'):
				f.write('  sigma %f\n' % e['sigma'])

		# Write out noise blocks.
		for nName in d['noise'].keys():
			n = d['noise'][nName]
			f.write('\nmodel noise %s\n' % nName)
			if self.isNumber(n, 'B'):
				f.write('  B %f\n' % n['B'])
			if self.isWord(n, 'env'):
				f.write('  env %s\n' % n['env'])
			if self.isWord(n, 'f'):
				f.write('  f %s\n' % n['f'])
			if self.isNumber(n, 'NF'):
				f.write('  NF %f\n' % n['NF'])
			if self.isNumber(n, 'T'):
				f.write('  T %f\n' % n['T'])

		# Write out stations blocks.
		for sName in d['station'].keys():
			s = d['station'][sName]
			f.write('\nmodel station %s\n' % sName)
			if self.isWord(s, 'ant'):
				f.write('  ant %s\n' % s['ant'])
			if self.isWord(s, 'pol'):
				f.write('  pol %s\n' % s['pol'])
			if self.isNumber(s, 'az'):
				f.write('  az %f\n' % np.degrees(s['az']))
			if self.isNumber(s, 'el'):
				f.write('  el %f\n' % np.degrees(s['el']))
			if self.isNumber(s, 'hgt'):
				f.write('  hgt %f\n' % s['hgt'])
			if self.isNumber(s, 'lat'):
				f.write('  lat %f\n' % np.degrees(s['lat']))
			if self.isNumber(s, 'lon'):
				f.write('  lon %f\n' % np.degrees(s['lon']))
			if self.isWord(s, 'env'):
				f.write('  env %s\n' % s['env'])
			if self.isWord(s, 'noise'):
				f.write('  noise %s\n' % s['noise'])
			if self.isWord(s, 'sys'):
				f.write('  sys %.1f\n' % s['sys'])
			if self.isWord(s, 'site'):
				f.write('  site %s\n' % s['site'])
			if self.isWord(s, 'tx'):
				f.write('  tx %.1f\n' % s['tx'])
				f.write('  service %d\n' % s['service'])

		f.close()

	def handleRbArea(self):
		'''ITM area mode is chosen, so gray out GUI options related to
		ITM profile mode.
		'''

		global d, u
		u.cbSiting.setEnabled(True)
		u.leDh.setEnabled(True)
		u.leLoc.setEnabled(True)
		d['rfl']['mode'] = 'area'
		return

	def handleRbCoverage(self):
		global d, u
		u.cbUnits.setEnabled(True)
		u.leGrid.setEnabled(True)
		u.leRange.setEnabled(True)
		u.leMap.setEnabled(True)
		u.leLevels.setEnabled(True)
		d['rfl']['output'] = 'coverage'

	def handleRbLink(self):
		global d, u
		u.cbUnits.setEnabled(False)
		u.leGrid.setEnabled(False)
		u.leRange.setEnabled(False)
		u.leMap.setEnabled(False)
		u.leLevels.setEnabled(False)
		d['rfl']['output'] = 'csv'

	def handleRbProfile(self):
		'''ITM profile mode is chosen, so gray out GUI options related to
		ITM area mode.
		'''

		global d, u
		u.cbSiting.setEnabled(False)
		u.leDh.setEnabled(False)
		u.leLoc.setEnabled(False)
		if d is None:
			return
		d['rfl']['mode'] = 'profile'

	def handleRbRx(self):
		global u
		u.leTx.setEnabled(False)
		u.cbService.setEnabled(False)

	def handleRbTx(self):
		global u
		u.leTx.setEnabled(True)
		u.cbService.setEnabled(True)

	def isNumber(self, dictionary, key):
		if not key in dictionary.keys():
			return False
		n = dictionary[key]
		if type(n) is str:
			return False
		return np.isscalar(n)

	def isWord(self, dictionary, key):
		return key in dictionary.keys() and str(dictionary[key]).strip() != ''

